/*
    Copyright (C) 2000 DFN-FWL (http://www.fwl.dfn.de/fwl)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
/*
 * $Id: filter_html.c,v 1.33 2000/04/16 15:52:12 gellert Exp $
 *
 * $Log: filter_html.c,v $
 * Revision 1.33  2000/04/16 15:52:12  gellert
 * fixed some short lines (the end of them was left out while cut and
 * paste)...
 *
 * Revision 1.32  2000/04/16 15:27:40  gellert
 * output_string called reset_buffers() which could lead to negative string
 * lengths in calls to itself. Fixed and added some checks ;)
 *
 * Revision 1.31  2000/03/07 15:06:49  goldbach
 * added GPL info
 *
 * Revision 1.30  2000/03/02 16:00:31  goldbach
 * modified: get_next_tag(): if we're skipping parts of the document the search
 * for the next tag is modified to check that the following is a closing tag.
 * That way we're safely processing script code with arithmetic expressions.
 *
 * Revision 1.27  2000/02/23 17:45:22  gellert
 * If the end of a comment or tag is not in the inbuf, httpf
 * now memorizes, what is to be searched: The end of a comment
 * or the end of a tag. So the correct end of the comments
 * is now found even in this case.
 *
 * Revision 1.26  2000/02/23 15:25:32  gellert
 * Fixed a bug in the finding of the end of a comment. The test, if the
 * found end-of-comment was behind the end of the read buffer, failed.
 *
 * Revision 1.25  2000/02/16 17:07:27  goldbach
 * modified log messages (thread id was missing)
 *
 * Revision 1.24  2000/02/16 14:21:02  goldbach
 * removed function make_test_file() that created a file with just a big comment.
 *
 * Revision 1.23  2000/02/16 13:33:43  goldbach
 * replaced remaining calls of syslog >:)
 *
 * Revision 1.22  2000/02/01 17:02:22  gellert
 * Fixed a bug (if x >= y)  --> (if x == y)
 *
 * Revision 1.21  1999/08/19 13:39:35  gellert
 * small fix in the handling of comments (now we don't expect the
 * number of "-" at the end of a comment to be even).
 *
 * Revision 1.20  1999/08/05 14:39:47  goldbach
 * moved code from get_end_of_tag to extra function skip_comment().
 * Now httpf processes comments correctly. Hopefully :)
 *
 * Revision 1.19  1999/08/04 13:36:20  gellert
 * Just fixed a bug in the part that has to find the end of a comment.
 * This works now...
 *
 * Revision 1.18  1999/07/28 14:18:52  gellert
 * now the scanning of comments works correctly:
 * search for -- then find the next >
 *
 * Revision 1.17  1999/07/22 16:02:41  gellert
 * replaced strchr by strnchr
 *
 * Revision 1.16  1999/07/22 15:35:03  goldbach
 * yet again cleanup for logmsg()
 *
 * Revision 1.15  1999/07/22 14:36:02  gellert
 * improved tag scanning (find_end_of_tag), including nested quotes.
 *
 * Revision 1.14  1999/07/22 14:28:01  goldbach
 * minor cleanup for logmsg() (again)
 *
 * Revision 1.13  1999/07/16 12:49:05  goldbach
 * fixed 'bug' that mumbled about illegal inbuf offset.
 *
 * Revision 1.12  1999/07/16 09:48:49  goldbach
 * minor cleanup (warning)
 *
 * Revision 1.11  1999/07/15 20:11:10  gellert
 * removed a comment
 *
 * Revision 1.10  1999/07/15 19:59:48  gellert
 * get_end_of_tag was extended. it now handles comments
 * correctly: if the tag is a comment, the it searches
 * for the next "-->" otherwise it searches for the
 * next '>'.
 *
 * Revision 1.9  1999/07/14 09:50:04  goldbach
 * removed last strncpy.
 *
 * Revision 1.8  1999/06/30 13:25:44  goldbach
 * removed test dumps to stderr
 *
 * Revision 1.7  1999/06/29 12:20:01  gellert
 * just fiddled with the array boundaries to prevent
 * errors.
 *
 * Revision 1.6  1999/06/15 16:18:15  goldbach
 * bug: read_body_after_broken_tag() killed some contents always...
 *
 * Revision 1.5  1999/06/15 14:40:34  goldbach
 * if there are more than HTTP_BUFSIZE bytes between '<' and '>' this
 * so-called "tag" is thrown awaz silently. Well, okay... there's a log
 * message that tells you how many bytes are suppressed.
 *
 * This is quite handy when binaries are sent as text/html (like Apache
 * does it. Try this at home kids: give a binary the extension '.html'
 * and Apache will send it as text/html :(
 *
 * Revision 1.4  1999/06/09 20:33:53  gellert
 * new version. now the parsing of tags and attributes works.
 * and it still works if the client cancels the transmission
 * (stop button in browser).
 *
 * Revision 1.3  1999/05/27 16:47:59  goldbach
 * minor cleanup for logmsg()
 *
 * Revision 1.2  1999/05/27 11:29:35  gellert
 * well, now it does what we want. just small mistakes (eg end of
 * comments) are done.
 *
 * Revision 1.1  1999/05/12 16:07:11  goldbach
 * first version
 *
 *
 */


#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <syslog.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include "parse_tag.h"
#include "taghash.h"
#include "httpf.h"
#include "mythreads.h"
#include "misc.h"
#include "read.h"
#include "filter_html.h"

/* prototypes */
char *copy_pretag_contents(char *, TCT *);
char *get_next_tag(TCT *);
char *get_end_of_tag(char *,TCT *);
int handle_tag(char *, TCT *);
int handle_broken_tag(char *, TCT *);
int read_body_after_broken_tag(TCT *);
char *skip_comment(TCT*, char*);


/*
 * filter_html()
 *
 * Run over the inbuf and filter the HTML code we find.
 * After this function is called the output buffer is ready
 * for sending and the input buffer is ready for receiving.
 * Note: reset_buffers() may NOT be called afterwards --- buffer
 * offsets are set inside these functions.
 *
 * Returns:
 * -1 -- an error occurred
 * other -- everything's normal
 */
int filter_html(TCT *tct)
{    
    int result= 1;
    logmsg(LOG_INFO, "[%d] Filtering %d bytes of HTML code.", tct->tid, tct->inlen); 
    
    while(result == 1) {
        result = handle_tag(copy_pretag_contents(get_next_tag(tct),tct),tct);
	logmsg(LOG_DEBUG, "[%d] Processed %d of %d bytes in input buffer.",
	       tct->tid, tct->inoffset, tct->inlen);
    }
    /* and return the value of handle_tag. if this is -1, then the loop
       in copy_pkts must be ended (the client closed the connection)
     */
    return result;
}



/*
 * get_next_tag()
 *
 * Find next occurrence of < in inbuf+inoffset.
 */
char *get_next_tag(TCT *tct) {
    char *tag_pos;
    
    /* check if the buffer is processed */
    if (tct->inoffset==tct->inlen)
	return NO_MORE_TAG;
    
    /* check if there's some error (we may have processed too much) */
    if (tct->inoffset>tct->inlen) {
	logmsg(LOG_ERR, "[%d] Illegal inbuf offset (%d of %d), aborting.", 
	       tct->tid, tct->inoffset, tct->inlen);
	return NO_MORE_TAG;
    }
	
    logmsg(LOG_DEBUG, "[%d] Inbuf offset is %d.", tct->tid, tct->inoffset);
    
    tag_pos = strnchr(tct->inbuf+tct->inoffset, '<',
		      tct->inlen-tct->inoffset); /* find < */
    
    do {

	/*
	 * here we're using this implied logic (TODO: doc :))
	 * skipping is done between tags of *one* HTML element.
	 * FIXME: check config upon init when this works. If this
	 * code remains here we might as well use just *one* entry
	 * entry in the config file: "replskip" implies "replendskip..."
	 * This would also reduce the number of possible config. mistakes  ;)
	 *
	 * Skipping may be ended only when we find a "</" since
	 * it starts with an opening tag and ends with a closing tag.
	 * Thus, when finding a "<" we check if we're currently skipping
	 * contents. If so, we just keep on searching for the next tag introducer.
	 * The "<" just found is probably part of an arithmetic expression in script code...
	 */
	if (tag_pos==NULL) return NULL;
	
	if (tct->skip_flag==1) { /* skipping? */
	    if ((tct->inoffset+1<tct->inlen) && (tag_pos[1]=='/')) { /* still in buf and is there "</"? */
		/* do return */
		return tag_pos;
	    } else {
		/* don't return, continue searching */
		tct->inoffset = tag_pos-tct->inbuf;
		tag_pos = strnchr(tag_pos+1, '<',
				  tct->inlen-tct->inoffset); /* find < */
	
	    }
	} else {
	    /* just return the position of the new tag if we're not skipping */
	    return tag_pos;
	}
    } while (tag_pos);
    
    return tag_pos;
}


/*
 * copy_pretag_contents
 *
 * Copy contents from inbuf to outbuf. This function is called
 * to do a verbatim copy between tags.
 */
char *copy_pretag_contents(char *tag, TCT *tct) 
{
    /* 
     * copy inbuf from the place we stopped working before
     * up to the tag. If we didn't find a tag, copy the rest of the
     * buffer.
     */

    if (tag!=NO_MORE_TAG) {
	/* if we have a tag in the rest of the buffer */
	output_string(tct->inbuf+tct->inoffset,
		      tag-(tct->inbuf+tct->inoffset),
		      tct);
	
	/* tell the others where we are */
	tct->inoffset=tag-tct->inbuf;
	/* return the place we have to start working again. */
	return tag;    
    } else {
	logmsg(LOG_INFO, "[%d] No more tags in this buffer.", tct->tid);
	logmsg(LOG_DEBUG, "[%d] Am at offset %d.", tct->tid, tct->inoffset);
	
	if (tct->inlen-tct->inoffset>+0) {
	    /* copy the complete rest of the buffer */
	    output_string(tct->inbuf+tct->inoffset,
			  tct->inlen-tct->inoffset,
			  tct);
	} else {
		logmsg(LOG_INFO, "[%d] length of pretag contents is negative", tct->tid);
         	logmsg(LOG_INFO, "     inlen=%d, inoffset=%d", tct->inlen, tct->inoffset);
	}
	
	return 0; /* finished with buffer */
    }
}

/*
 * get_end_of_tag()
 *
 * Find next occurrence of > after tag location.
 * Returns:
 *   NULL -- no closing '>' in this buffer or
 *           comment skipped.
 */
char *get_end_of_tag(char *tag, TCT *tct) 
{
    char *t=tag; char *s;
    char buffer_end = 0;

    logmsg(LOG_DEBUG, "[%d] Tag at offset %d.", tct->tid, tag-tct->inbuf);

    if ((*(t+1)== '!') && (*(t+2) == '-') && (*(t+3) == '-')) {
	/* ok, the < was the start of a comment, skip it */
	tct->end_to_search=END_OF_COMMENT;
	return skip_comment(tct, tag);
    }

    /* the tag is not a comment, so now just find the closing > */
    tct->end_to_search=END_OF_TAG;
    s=find_end_of_tag(t, tct);
    if(s==NULL) logmsg(LOG_DEBUG, "[%d] No '>' in this buffer.", tct->tid);
    
    return s;
}


/* handle_tag: return values
** -1 = error, sending was impossible (this happens when the client
**	cancels the connection = stop-button in browser)
** 0  = everything ok, the buffer was fully processed
** 1  = everything ok, go on with tag-handling
*/
int handle_tag(char *tag, TCT *tct) 
{
    char *eot;

    /* send and return NULL if we've finished this buffer */
    if (tag==NULL) {
	logmsg(LOG_INFO, "[%d] Buffer processed, resetting.", tct->tid);
	logmsg(LOG_INFO, "[%d] Sending %d bytes.", tct->tid, tct->outoffset);
	if (send(tct->insocket, tct->outbuf, tct->outoffset, 0)<0) {
	    logmsg(LOG_INFO, "[%d] Sending on client socket failed with errno %d.",
		   tct->tid, errno);
	    reset_buffers(tct);
	    return -1;
	    }
	reset_buffers(tct);
	return 0;
    }

    /* find end of tag */
    eot=get_end_of_tag(tag, tct);

    /* handle this broken tag and leave */
    if (eot==NULL) return handle_broken_tag(tag, tct);
    
    /* now we know we have a decent tag  */
    logmsg(LOG_DEBUG, "[%d] Processing tag.", tct->tid);

    eot[0]=0; /* replace '>' in inbuf with null character */
    
    /* parse it! */
    parse_tag(tag+1, tct);
    
    tct->inoffset=eot-tct->inbuf+1;
    logmsg(LOG_DEBUG, "[%d] After parse tag: inoffset=%d, outoffset=%d.",
	   tct->tid, tct->inoffset, tct->outoffset);
    
    return 1;
}

int handle_broken_tag(char *tag, TCT *tct) 
{
    int len;
    
    logmsg(LOG_DEBUG, "[%d] Handling broken tag.", tct->tid);

    /* send the outbuf so far */
    logmsg(LOG_INFO, "[%d] Broken Tag. outlen: %i", tct->tid, tct->outlen);
    if (send(tct->insocket, tct->outbuf, tct->outlen, 0)<0) {
	  logmsg(LOG_INFO, "[%d] Sending on client socket failed with errno %d.",
		 tct->tid, errno);
	  reset_buffers(tct);
	  return -1;
	  }

    len=tct->inbuf+tct->inlen-tag;
    logmsg(LOG_DEBUG, "[%d] Copying %d bytes from offset %d to the beginning of inbuf",
	   tct->tid, len, tag-tct->inbuf);

    /* copy inbuf from tag till end to start of inbuf */
    strncpy(tct->inbuf, tag, len);

    /* init. offsets */
    reset_buffers(tct);
    
    /* we already have a (small) chunk of valid data in the buffer */
    tct->inlen = len;

    /* read next block */
    return read_body_after_broken_tag(tct);
    
    /*
     * now the part containing the broken tag is at the beginning
     * of the inbuf. it's vital *not* to call reset_buffers()
     * afterwards since we set the inlen attribute.
     */
    
}

/*
 * read_body_after_broken_tag
 *
 * Read body until we find a closing brace. If the amount of data between
 * the braces '<' and '>' exceeds the buffer size we silently throw the buffer
 * contents away.
 * Returns:
 *  0 -- ok
 * -1 error whilst reading
 */
int read_body_after_broken_tag(TCT *tct)
{
    int quit=0; /* 1 -- ok, -1 -- error whilst reading */
    char *pos;
    int len, total_size_skipped=0;
    int reads=0; /* #of calls to read_server_body */
    
    logmsg(LOG_DEBUG, "[%d] Entering read_body_after_broken_tag()", tct->tid);
    
    while(quit==NULL) {
	reads+=1;
	logmsg(LOG_DEBUG, "[%d] Reading next server body", tct->tid);
	if (read_server_body(tct)!=0) {
	    /* leave if reading failed */
	    logmsg(LOG_INFO, "[%d] Tag incomplete, %d bytes ignored.", tct->tid, total_size_skipped);
	    return -1;
	}
	
	if (tct->end_to_search == END_OF_TAG) {
	    logmsg(LOG_DEBUG, "[%d] Looking for '>'", tct->tid);
	    /* look for '>' */
	    pos = strnchr(tct->inbuf, '>', tct->inlen);
	} else {
	    logmsg(LOG_DEBUG, "[%d] Looking for end of comment", tct->tid);
	    pos = skip_comment(tct, tct->inbuf);
	}
	
	if (pos!=NULL) {
	    /* we found it */
	    logmsg(LOG_DEBUG, "[%d] Found it in read loop %d at offset %d", tct->tid, reads, pos-tct->inbuf);

	    /*
	     * If we read more often than once from the server socket
	     * our "tag" has a size of more than HTTP_BUFSIZE bytes.
	     * In this case we copy the contents we just read to the
	     * beginning of the buffer (and ignore the bytes between
	     * the '<' and the '>').
	     */
	    if (reads!=1) {
		/* we read more than once */
		len = tct->inbuf+tct->inlen-pos-1;
		memcpy(tct->inbuf, pos+1, len); /* copy stuff behind the '>' */
		logmsg(LOG_DEBUG, "[%d] Copied %d bytes from offset %d", tct->tid, len, (int)(pos-tct->inbuf));
		total_size_skipped += tct->inlen-len-2; /* inc. #of bytes skipped. -2 are the braces :) */
		tct->inlen = len; /* set new buffer length */
	    }
	    
	    quit=1; /* quit loop */
	} else {
	    logmsg(LOG_DEBUG, "[%d] Ignoring body", tct->tid);
	    total_size_skipped += tct->inlen; /* inc. #of bytes skipped */
	    /* didn't find it: ignore contents */
	    reset_buffers(tct);
	}
    }

    /* log msg if something was skipped */
    if (reads!=1) {
	logmsg(LOG_INFO, "[%d] Ignored tag of size %d", tct->tid, total_size_skipped);
    }

    return 1;   
}


/*
 * skip_comment
 *
 * Skips HTML comment.
 * Input:
 *   thread context, pointer to current tag.
 * Returns:
 *    pointer to end of comment
 */
char *skip_comment(TCT *tct, char *tag) {
    int found=0;
    char *hyphen, *t;
    int rest_buf_len;
    int buffer_end=0;
    
    t=tag+4;	/* skip '<!--'  */

    logmsg(LOG_DEBUG, "[%d] Found comment at offset %d.", tct->tid, tag-tct->inbuf);
    while (!found && !buffer_end) {
/*        rest_buf_len = t - tag + tct->inlen - tct->inoffset - 4;  these are the four bytes we skipped (t=t+4) */
	rest_buf_len = tct->inlen-(t-tct->inbuf);
	hyphen=strnchr(t, '-', rest_buf_len); /* find first '-' */

        if (hyphen != NULL) { /* found it? */
	    /* now the rest to '-->' */
	    if ((*(hyphen+1)=='-')) {
	      /* this is the end of a comment */
		t=hyphen+2; /* point right behind 2nd '-' */
		logmsg(LOG_DEBUG, "[%d] Found closing -- at offset %d.", tct->tid,t-tct->inbuf);
		t = skip_leading_whitespace(t);
		/* check if we're in illegal region */
		if (t>(tct->inbuf+tct->inlen)) {
		    buffer_end=1;		    
		} else {
		    if (t[0]=='>') {
			/* found closing '>' */
			found=1;
			return t;
		    } else {
			logmsg(LOG_ERR, "[%d] Comment was closed but continues at offset %d.",
			       tct->tid, t-tct->inbuf);
			t-=1; /* step back */
		    }
		    /* t points now to the char right behind the second '-' */
		}
	    } else {
		/* this was just an ordinary '-' in a comment */
		t=hyphen+1; /* step behind the lonely '-' (NB: the +1 is important :)) */
		hyphen=NULL; /* we didn't find the closing part, so if
				  the buffers ends now, then return NULL
				  (not the position of the '-' we found */
		if (t > tct->inbuf + tct->inlen) {
		    /* end of buffer reached */
		    logmsg(LOG_DEBUG, "[%d] Broken comment: rest_buf_len %d.", tct->tid, rest_buf_len);
		    buffer_end = 1;
		}
	    }
        } else {
	    /* no '-' found */
	    logmsg(LOG_DEBUG, "[%d] No closing '-->' in this buffer (ends with '%8s').",
             tct->tid, tct->inbuf+tct->inlen-8);
	    buffer_end = 1;
        }
    } /* end of while */
    return NULL;
}

