/*
    Copyright (C) 2000 DFN-FWL (http://www.fwl.dfn.de/fwl)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
/*
 * $Id: ftype_check.c,v 1.5 2000/03/07 15:06:50 goldbach Exp $
 *
 * $Log: ftype_check.c,v $
 * Revision 1.5  2000/03/07 15:06:50  goldbach
 * added GPL info
 *
 * Revision 1.4  1999/08/04 11:41:43  gellert
 * Changed the type check. Now there are two kinds of entries in httpf.cfg:
 * 1) ftype mimetype offset pattern: If the given mimetype matches, than
 *    it is checked whether the block contains pattern at offset.
 * 2) ftype mimetype: If the given mimetype matches, no further check follows.
 *    This makes sense for types like text/plain...
 * These changes required the change of the threadcontext, there is now a
 * character string containing the given mimetype of the server response.
 *
 * Revision 1.3  1999/07/28 14:19:29  gellert
 * some changes to the calls of logmsg
 *
 * Revision 1.2  1999/06/09 20:33:53  gellert
 * new version. now the parsing of tags and attributes works.
 * and it still works if the client cancels the transmission
 * (stop button in browser).
 *
 * Revision 1.1  1999/05/27 11:32:33  gellert
 * check type of files (eg gif, jpeg, ps, pdf)
 *
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <syslog.h>
#include <ctype.h>
#include "ftype_check.h"
#include "misc.h"

struct type_list {
    struct	type_list *next;	/* next entry */
    char	*pattern;		/* the pattern to match */
    char	*mime_type;		/* the coresponding mime-type */
    uint	offset;			/* where is the pattern */
    uint	pattern_length;		/* length of the pattern */
};

static struct type_list *mime_types = NULL;
/* static struct type_list mime_types = {NULL, NULL, NULL, 0}; */

int type_insert(char *pattern, char *type, int offset, int length) {

 struct	type_list *actual;	/* pointer to the actual item */
 struct type_list *new_one;	/* pointer to new generated item */

 new_one = (struct type_list *) malloc(sizeof(*new_one));
 if (new_one == NULL) {
     return (-1);
     }
 actual = mime_types;
 if (mime_types == NULL) {
   /* this is the first call to type_insert */
   /* so set the start_pointer to the first node */
   mime_types = new_one;
 } else {
    /* this is not the first call to type_insert */
   /* so go to the end of the list */
   while (actual->next != NULL) {
     actual = actual->next;
     }
   actual->next = new_one;
 } /* end of else */

 /* now we just fill in the values */
 new_one->next = NULL;
 new_one->mime_type = strdup(type);
 if (pattern != NULL) {
 logmsg(LOG_DEBUG, "Now inserting item! With pattern.");
   new_one->pattern = strdup(pattern);
   new_one->offset = offset;
   new_one->pattern_length = length;
   logmsg(LOG_DEBUG, "Scanned file type: mtype %s :: pattern %s :: offset %u.\n",
	new_one->mime_type, new_one->pattern, new_one->offset);
 } else {
   logmsg(LOG_DEBUG, "Now inserting item! No pattern.");
   new_one->pattern = NULL;
   new_one->offset = 0;
   new_one->pattern_length = 0;
   logmsg(LOG_DEBUG, "Scanned file type: mtype %s.\n",
	new_one->mime_type);
 }
 return (0);
}

int type_check(TCT *tct) {
    char *begin=tct->inbuf+tct->inoffset;
    int length=tct->inlen-tct->inoffset;
    char *mimetype=tct->mime_type;
    
    struct	type_list *actual;	/* pointer to the actual item */
    int type_found = 0;
    int done = 0;

    actual = mime_types;	/* set pointer to first list item */
    while (!done) {
      type_found = 0;
      while ((!type_found) && (actual != NULL)) {
	if (strncmp(mimetype, actual->mime_type, strlen(mimetype))==0) {
	    /* ok, we found the right mimetype */
	    type_found = 1;
	} else {
	    actual = actual->next;
	}
      } /* end of while */
    
      if (actual != NULL) {
	/* this is a pointer to the matching type */
	logmsg(LOG_NOTICE, "[%d] Found matching mimetype for %s.", tct->tid, mimetype);
	if (actual->pattern != NULL) {
	    /* do additional check! */
	    logmsg(LOG_DEBUG, "[%d] Doing configured check on mimetype %s", tct->tid, mimetype);
	    if (actual->offset+actual->pattern_length > length) {
		/* the read block is not long enough, the pattern is not in there */
		logmsg(LOG_ERR, "[%d] Couldn't scan for pattern, block too short.", tct->tid);
		logmsg(LOG_ERR, "[%d] MimeType: %s, DataLength: %i, PatternOffset: %i, Pattern: %s\n",
		       tct->tid, actual->mime_type, length, actual->offset, actual->pattern);
		return 1;
	    } else {
		logmsg(LOG_DEBUG, "[%d] Checking '%s' with '%s'.", tct->tid, actual->pattern,begin+actual->offset);
		if (bcmp(actual->pattern, begin+actual->offset, actual->pattern_length)==0) {
		    return 0;
		} else {
		    logmsg(LOG_DEBUG, "[%d] Types didn't match. Trying next.", tct->tid);
		    actual = actual->next;
		    if (actual!=NULL) {
		      logmsg(LOG_DEBUG, "[%d] Next pattern will be: %s", tct->tid, actual->pattern);
		    }
		}
	    }
	} else {
	    /* no additional check was configured in httpf.cfg */
	    return 0;
	}
      } else {
        done = 1;
      }
  }  
    logmsg(LOG_ERR, "[%d] Unconfigured mimetype found: %s", tct->tid, mimetype);
    return 1;   
}

