/*
    Copyright (C) 2000 DFN-FWL (http://www.fwl.dfn.de/fwl)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
/*
 * $Id: mythreads.h,v 1.13 2000/03/07 15:06:53 goldbach Exp $
 *
 * $Log: mythreads.h,v $
 * Revision 1.13  2000/03/07 15:06:53  goldbach
 * added GPL info
 *
 * Revision 1.12  2000/02/23 17:45:22  gellert
 * If the end of a comment or tag is not in the inbuf, httpf
 * now memorizes, what is to be searched: The end of a comment
 * or the end of a tag. So the correct end of the comments
 * is now found even in this case.
 *
 * Revision 1.11  1999/12/01 17:53:02  gellert
 * bug-fix: field_name was a static pointer (used by all the threads!). ARGH!
 *
 * Revision 1.10  1999/11/26 10:47:08  goldbach
 * New actions introduced: replskip and replendskip skip sections of documents, i.e. data won\'t be sent to the client.
 *
 * Revision 1.9  1999/08/04 11:41:45  gellert
 * Changed the type check. Now there are two kinds of entries in httpf.cfg:
 * 1) ftype mimetype offset pattern: If the given mimetype matches, than
 *    it is checked whether the block contains pattern at offset.
 * 2) ftype mimetype: If the given mimetype matches, no further check follows.
 *    This makes sense for types like text/plain...
 * These changes required the change of the threadcontext, there is now a
 * character string containing the given mimetype of the server response.
 *
 * Revision 1.8  1999/05/12 16:04:23  goldbach
 * major clean-up session
 *
 * Revision 1.7  1999/03/30 12:30:07  goldbach
 * 1st surfable version
 *
 * Revision 1.6  1999/03/02 11:39:08  goldbach
 * added length information for buffers.
 *
 * Revision 1.5  1999/02/25 13:37:21  goldbach
 * added fields from the request line
 *
 * Revision 1.4  1999/02/18 16:46:48  goldbach
 * added loglevel
 *
 * Revision 1.3  1999/02/11 15:08:37  goldbach
 * Sockets fr aktuelle Verbindung eingefgt
 *
 * Revision 1.2  1999/02/11 14:30:55  goldbach
 * Puffer fuer Ein-, Ausgabe und Fehlerverarbeitung sowie Offsets
 * hinzugefuegt.
 *
 */

#ifndef __MYTHREADS__
#define __MYTHREADS__
#define WORKING 1
#define ERROR -1
#define READY 0
#define SCHEDULED 2

/* length of the reason phrase */
#define LEN_REASON 40

/* what are we searching for? the end of a tag or the end of a comment? */
#define	END_OF_TAG 0
#define	END_OF_COMMENT 1

#include <semaphore.h>

typedef struct thread_context {
    int tid; /* eigene Thread ID, != pthread_self() !!!! */
    sem_t wait;
    int stat; /* <0 = error; == 0 = ready; > 0 = working */
    int sockfd;
    /* ab hier aufgabenspezifischer Krams */
    int ServerSockSize; 
    int ClientSockSize;
    int CopySize;
    /*
     * buffers for input, output and error handling.
     * NB: inbuf will be modified (e.g. insert null character
     * for CR, LF).
     */
    char *inbuf;
    char *outbuf;
    char *errbuf;
    /* buffer offsets. incremented whilst processing. */
    int inoffset;
    int outoffset;
    int erroffset;
    /* valid len */
    int inlen; /* bytes read */
    int outlen; /* bytes to be written */
    int errlen; /* bytes to be written */
    /* sockets for this connection */
    int insocket;
    int outsocket;
    /* log level */
    int loglevel;

    /* request line: http method */
    int method;
    /* request line: HTTP/major.minor */
    int version_major;
    int version_minor;
    /*
     * request line: URI
     * this will be somewhere at the start of inbuf
     */
    char *uri;
    /* response line: status */
    int response_status;
    /* response_line: reason phrase */
    char reason[LEN_REASON];
    /* MIME type id from server response (subrange) */
    char *mime_type;
    /* field_name: points to the actual fieldname in the http header
       (used by copy_headerline in analyze_header.c) */
    char *field_name;
    /* flag for skipping parts of documents */
    int skip_flag;
    /* what are we searching for: the end of a tag or the end of
       a comment? */
    int end_to_search;
} TCT;


#endif


